{{
*************************************
* Prop-BIOS module
*************************************

Concept of PropBIOS services is to provide an interface that can be
used by any application that is aware of the PropBIOS services API.

Basic Features (PhaseI):
* Selected BIOS services built with a pre-processor.
* The core API is language neutral.
* Wrappers provide interface to user's programming language.
* Some underlying code written in .spin for size optimization/convenience.
* Hardware device configuration defined by user specified PinDefs.spin file.

Extended Features (PhaseII):
* Underlying code migratable to .asm for size/speed optimization.

*************************************
* Prop-BIOS modular framework
*************************************

Module can be compiled with Propeller IDE before or after running cpp pre-processor.
The cpp pre-processor is available in linux or in windows with cygwin package.

The cpp pre-processor in CygWin is available from: http://www.cygwin.com/

A typical make.bat file might look like this:

@rem # #######################################################
@rem #
@rem # PropBIOS make file: requires a pre-processor like cpp
@rem #
@rem # #######################################################
@rem #
@rem # Services are selected using the FEATURES variable i.e.
@rem # set FEATURES=-DUSE_PINDEFS -DUSE_TV
@rem #
@rem # Not all services are currently available, but may include
@rem #
@rem # USE_PinDefs_<user-name> ... potentially
@rem # USE_TV
@rem # USE_VGA
@rem # USE_SERIAL
@rem # USE_MOUSE
@rem # USE_KEYBOARD
@rem # USE_EEPROM
@rem # USE_SDCARD
@rem #
@rem # #######################################################

@rem # make vars

set FEATURES=-DUSE_TV -DUSE_SERIAL -DUSE_MOUSE -DUSE_KEYBOARD
@rem set FEATURES=-DUSE_TV -DUSE_VGA -DUSE_SERIAL

set CPP=cpp

@rem # make command

%CPP% %FEATURES% -P BIOS.spin BIOS-select.spin

@rem # sanity check
cat BIOS-select.spin

@rem # #######################################################
@rem # end of make.bat
@rem # #######################################################

}}

OBJ
  pdef  : "PinDefs"
  bc    : "BIOS_Interface"
  crc   : "CRC16"
  sconv : "Itoatoi"
  sf    : "snprintf"

CON ' PinDefs
   _clkmode  = pdef#_clkmode
   _xinfreq  = pdef#_xinfreq

  LF = $A
  CR = $D
CON ' USE_TV
{
#ifdef USE_TV
}
OBJ
  tv            : "TV_Text2"
{
#endif
}
CON ' USE_VGA
{
#ifdef USE_VGA
}
OBJ
  vga           : "VGA_Text2"
{
#endif
}
CON ' USE_SERIAL
{
#ifdef USE_SERIAL
}
OBJ
  fds           : "FullDuplexSerial"
{
#endif
}
CON ' USE_MOUSE
{
#ifdef USE_MOUSE
}
OBJ
  mouse         : "Mouse"
{
#endif
}
CON ' USE_KEYBOARD
{
#ifdef USE_KEYBOARD
}
OBJ
  keybd         : "Keyboard"
{
#endif
}

CON ' USE_EEPROM
{
#ifdef USE_EEPROM
}
OBJ
  eep        : "EEPROM"
{
#endif
}
CON ' USE_SDCARD
{
#ifdef USE_SDCARD
}
OBJ
  sdcard        : "fsrw"
{
#endif
}

VAR
  long cog
  long stack[100]     ' probably too big
  byte cbuf[33]       ' for itoatoi


PUB start
{{
Start the BIOS cog for serving non-linked client requests.
}}
' Initialize devices
  initDevices

  long[bc#BIOS_CMD_BASE] := bc#BIOS_CONFIG_INIT
  cog := cognew(spin,@stack) + 1
  long[bc#BIOS_CMD_BASE] := 0
  result := cog

PUB stop
{{
Stop the cog
}}
  if cog
    cogstop(cog~ - 1)


CON
  #0
  TV_BASEPIN_BYTE    
  VGA_BASEPIN_BYTE    
  NTSCPAL_BYTE    
  SDCARD_BASEPIN_BYTE    
  MOUSE_DAT_PIN_BYTE    
  MOUSE_CLK_PIN_BYTE    
  KEYBD_DAT_PIN_BYTE    
  KEYBD_CLK_PIN_BYTE    
  #0
  ENET_CS_PIN_BYTE    
  ENET_SCK_PIN_BYTE    
  ENET_SI_PIN_BYTE    
  ENET_SO_PIN_BYTE    
  ENET_INT_PIN_BYTE    

DAT
  DefaultProduct_str    byte "Prop-BIOS", 0
  TV_BASEPIN_CFG_STR    byte "TV  Basepin [#/n]", 0
  VGA_BASEPIN_CFG_STR   byte "VGA Basepin [#/n]", 0
  NTSCPAL_CFG_STR       byte "NTSC? [Y/n]", 0
  SDCARD_BASEPIN_CFG_STR byte "SD Card Basepin [#/n]", 0
  MOUSE_DAT_PIN_CFG_STR byte "Mouse DAT pin [#/n]", 0
  MOUSE_CLK_PIN_CFG_STR byte "Mouse CLK pin [#/n]", 0
  KEYBD_DAT_PIN_CFG_STR byte "Keyboard DAT pin [#/n]", 0
  KEYBD_CLK_PIN_CFG_STR byte "Keyboard CLK pin [#/n]", 0
  ENET_CS_PIN_CFG_STR   byte "ENET CS  pin [#/n]", 0
  ENET_SCK_PIN_CFG_STR  byte "ENET SCK pin [#/n]", 0
  ENET_SI_PIN_CFG_STR   byte "ENET SI  pin [#/n]", 0
  ENET_SO_PIN_CFG_STR   byte "ENET SO  pin [#/n]", 0
  ENET_INT_PIN_CFG_STR  byte "ENET INT pin [#/n]", 0
   
CON

  BOOT_MODE_VIRGIN            = $00000000 ' if no bits, first boot 
  BOOT_MODE_DEBUG             = $00000001 ' a mask: if 0, no debug boot
  BOOT_MODE_RECONFIG          = $00000002 ' a mask: if 0, normal boot
  BOOT_MODE_BOOTC             = $0000000C ' a mask: if 0, boot spin
  BOOT_MODE_STDOUT_TV         = $00000010 ' an enumeration: if 0, use serial 
  BOOT_MODE_STDOUT_VGA        = $00000020 ' an enumeration:
  BOOT_MODE_STDOUT_OTHER      = $00000030 ' an enumeration:
  BOOT_MODE_STDOUT_MASK       = $000000F0 ' a mask: 
  BOOT_MODE_STDERR_TV         = $00000100 ' an enumeration: if 0, use serial
  BOOT_MODE_STDERR_VGA        = $00000200 ' an enumeration: 
  BOOT_MODE_STDERR_OTHER      = $00000300 ' an enumeration: 
  BOOT_MODE_STDERR_MASK       = $00000F00 ' a mask: 
  BOOT_MODE_STDIN_KEYBD       = $00000100 ' an enumeration: if 0, use serial
  BOOT_MODE_STDIN_OTHER       = $00000200 ' an enumeration: 
  BOOT_MODE_STDIN_MASK        = $00000F00 ' a mask: 
  BOOT_MODE_DHCP              = $00001000 ' a mask: if 0, use fixed ip
  BOOT_MODE_SDBOOT            = $00010000 ' a mask: if 0, just boot embedded code
  BOOT_MODE_NETBOOT           = $00020000 ' a mask: if 0, just boot embedded code
  BOOT_MODE_BOOTMASK          = $000F0000 ' a mask: if 0, just boot embedded code
   

VAR ' Config Version 1.00 overlay ....

  long cfg_size                ' keep as size-4 to exclude crc16      
  long cfg_version             ' Version number represented as $HHdd H is header version, d is data version
  long cfg_chipid              ' 
  long cfg_boot_mode           ' 
  long cfg_pingrp1[2]          ' offset $10: tv base, vga base, ntsc/pal, sdcard base, mouse dat, mouse clk, kb dat, kb clk
  long cfg_pingrp2[2]          ' ethernet pins: i_cs, i_sck, i_si, i_so, i_int
  long cfg_pingrpRSV[4]        '  
  long cfg_bootOk_cnt          ' 
  long cfg_bootFail_cnt        ' 
  long cfg_bootCfgErr_cnt      ' 
  long cfg_bootMisc_cnt        ' 
  long cfg_prodname[8]         ' 32 chars ... product name or copyright  
  long cfg_userid[4]           ' 16 chars 
  long cfg_passwd[4]           ' 16 chars 
  long cfg_hostid[4]           ' 16 chars 
  long cfg_ipv4_address        '   
  long cfg_ipv4_netmask        ' 
  long cfg_ipv4_gateway        ' 
  long cfg_ipv4_dns            ' 
  long cfg_macaddr[2]          ' 8 bytes ... only need 6 
  long cfg_ipv4_vlan           ' 
  long cfg_rsrv[6]             ' 
  long cfg_bootp_fname[4]      ' 16 chars     
  long cfg_sdboot_fname[4]     ' 16 chars
  long cfg_string_data[10]     ' 40 chars
  long cfg_crc16               '

  
PRI initDevices | badcrc, rc

'SetClk

' put bios config stuff here

{
#ifndef USE_EEPROM
}

' initialize output devices based on compile-time definitions
{
#ifdef USE_VGA
}
  vga.start(pdef#VGA1)
  if bc#BIOS_STDOUT_DEV == bc#BIOS_STDOUT_VGA
    bc.setConsole(bc#BIOS_STDOUT, bc#BIOS_VGA)
  if bc#BIOS_STDERR_DEV == bc#BIOS_STDERR_VGA
    bc.setConsole(bc#BIOS_STDERR, bc#BIOS_VGA)
{
#endif
}
{
#ifdef USE_TV
}
  tv.start(pdef#TV_DAC1,pdef#NTSC=="Y")
  if bc#BIOS_STDOUT_DEV == bc#BIOS_STDOUT_TV
    bc.setConsole(bc#BIOS_STDOUT, bc#BIOS_TV)
  if bc#BIOS_STDERR_DEV == bc#BIOS_STDERR_TV
    bc.setConsole(bc#BIOS_STDERR, bc#BIOS_TV)
{
#endif
}
{
#ifdef USE_MOUSE
}
  mouse.start(pdef#MOUSE1_DAT, pdef#MOUSE1_CLK)
{
#endif
}
{
#ifdef USE_KEYBOARD
}
  keybd.start(pdef#KEYBOARD1_DAT, pdef#KEYBOARD1_CLK)
  if bc#BIOS_STDIN_DEV == bc#BIOS_STDIN_KEY
    bc.setConsole(bc#BIOS_STDIN,  bc#BIOS_KEYBOARD)
{
#endif
}


{
#ifdef USE_SERIAL
}
  fds.start(pdef#SERIAL1_RX, pdef#SERIAL1_TX, pdef#SERIAL1_MODE, pdef#SERIAL1_BAUD)
  if bc#BIOS_STDIN_DEV == bc#BIOS_STDIN_SER
    bc.setConsole(bc#BIOS_STDIN,  bc#BIOS_FDSERIAL)
  if bc#BIOS_STDOUT_DEV == bc#BIOS_STDOUT_SER
    bc.setConsole(bc#BIOS_STDOUT, bc#BIOS_FDSERIAL)
  if bc#BIOS_STDERR_DEV == bc#BIOS_STDERR_SER
    bc.setConsole(bc#BIOS_STDERR, bc#BIOS_FDSERIAL)
{
#endif
}


{
#else
// USE_EEPROM
}

  eep.start(pdef#EEPROM1_SCL, pdef#EEPROM1_DAT)

  badcrc := getEEconfig(@cfg_size, @cfg_crc16-@cfg_size+4)

  if badcrc == 0
  
    initEEconfig
     
    print(string(CR,"Booting "))

    if strsize(@cfg_prodname)    
      printf(string("%s "),parm(@cfg_prodname))
          
    puts(string("..."))
     
    if (cfg_boot_mode & BOOT_MODE_RECONFIG)
      biosConfig(true)
     
  badcrc := getEEconfig(@cfg_size, @cfg_crc16-@cfg_size+4)

  if badcrc          ' if badcrc force config routine
    biosConfig(true)
  else               ' show config with menu on ESC detected in 3 seconds
    biosConfig(false)

  ' try to boot from sdcard or network if available
  if (cfg_boot_mode & BOOT_MODE_BOOTMASK)
    rc := \boot
    if rc
      printf(string("\nBoot failed. Return code: %d\n"), rc)
    repeat
  
{
#endif
}


{
#ifdef USE_EEPROM
}

PRI newEEconfig | ii, len, adj
  adj := bc#BIOS_DBASE_ADDR - @cfg_size
  puts(string(CR,"Preparing new config ..."))
  waitcnt(clkfreq+cnt)

  len := @cfg_crc16 - @cfg_size
  bytefill(@cfg_size, 0, len)

  cfg_size         := @cfg_crc16 - @cfg_size
  cfg_version      := $0000     ' 0 for now, $101 after published
  cfg_chipid       := chipver
  bytemove(@cfg_prodname, @DefaultProduct_str, strsize(@DefaultProduct_str)+1) ' copy null too
  
  byte[@cfg_pingrp1 + TV_BASEPIN_BYTE]                  := pdef#TV_DAC1
  byte[@cfg_pingrp1 + VGA_BASEPIN_BYTE]                 := pdef#VGA1
  byte[@cfg_pingrp1 + NTSCPAL_BYTE]                     := pdef#NTSC
  byte[@cfg_pingrp1 + SDCARD_BASEPIN_BYTE]              := pdef#FSRW1_DO
  byte[@cfg_pingrp1 + MOUSE_DAT_PIN_BYTE]               := pdef#MOUSE1_DAT
  byte[@cfg_pingrp1 + MOUSE_CLK_PIN_BYTE]               := pdef#MOUSE1_CLK
  byte[@cfg_pingrp1 + KEYBD_DAT_PIN_BYTE]               := pdef#KEYBOARD1_DAT
  byte[@cfg_pingrp1 + KEYBD_CLK_PIN_BYTE]               := pdef#KEYBOARD1_CLK

{
#ifdef USE_ENC28J60
}
  byte[@cfg_pingrp2 + ENET_CS_PIN_BYTE]                 := pdef#ENC1_CS
  byte[@cfg_pingrp2 + ENET_SCK_PIN_BYTE]                := pdef#ENC1_SCK
  byte[@cfg_pingrp2 + ENET_SI_PIN_BYTE]                 := pdef#ENC1_SI
  byte[@cfg_pingrp2 + ENET_SO_PIN_BYTE]                 := pdef#ENC1_SO
  byte[@cfg_pingrp2 + ENET_INT_PIN_BYTE]                := pdef#ENC1_INT
{
#endif
}
  
  cfg_crc16 := crc.get(@cfg_size, cfg_size,crc#CRC_INIT)
  puts(string(CR,"Use 'C' to config the'q' to quit and save Config ..."))

{
  printStdout(string("Saving new config with CRC "))
  printStdout(sconv.itoab(cfg_crc16,@cbuf,16))
  saveEEconfig(@cfg_size, cfg_size+4)
}
  

PRI initEEconfig

  print(string(CR,"Initializing devices with config-spec ...",CR))

' initialize output devices, re-init after config load
{
#ifdef USE_VGA
}
  if byte[@cfg_pingrp1+VGA_BASEPIN_BYTE] <> "n"
    vga.start(byte[@cfg_pingrp1+VGA_BASEPIN_BYTE])
    if (cfg_boot_mode & BOOT_MODE_STDOUT_MASK) == BOOT_MODE_STDOUT_VGA
      bc.setConsole(bc#BIOS_STDOUT, bc#BIOS_VGA)
    if (cfg_boot_mode & BOOT_MODE_STDERR_MASK) == BOOT_MODE_STDERR_VGA
      bc.setConsole(bc#BIOS_STDERR, bc#BIOS_VGA)
{
#endif
}
{
#ifdef USE_TV
}
  if byte[@cfg_pingrp1+TV_BASEPIN_BYTE] <> "n"
    tv.start(byte[@cfg_pingrp1+TV_BASEPIN_BYTE],byte[@cfg_pingrp1+NTSCPAL_BYTE])
    if (cfg_boot_mode & BOOT_MODE_STDOUT_MASK) == BOOT_MODE_STDOUT_TV
      bc.setConsole(bc#BIOS_STDOUT, bc#BIOS_TV)
    if (cfg_boot_mode & BOOT_MODE_STDERR_MASK) == BOOT_MODE_STDERR_TV
      bc.setConsole(bc#BIOS_STDERR, bc#BIOS_TV)
{
#endif
}
{
#ifdef USE_MOUSE
}
  if byte[@cfg_pingrp1+MOUSE_DAT_PIN_BYTE] <> "n" and byte[@cfg_pingrp1+MOUSE_CLK_PIN_BYTE] <> "n"
    mouse.start(byte[@cfg_pingrp1+MOUSE_DAT_PIN_BYTE], byte[@cfg_pingrp1+MOUSE_CLK_PIN_BYTE])
{
#endif
}
{
#ifdef USE_KEYBOARD
}
  if byte[@cfg_pingrp1+KEYBD_DAT_PIN_BYTE] <> "n" and byte[@cfg_pingrp1+KEYBD_CLK_PIN_BYTE] <> "n"
    keybd.start(byte[@cfg_pingrp1+KEYBD_DAT_PIN_BYTE], byte[@cfg_pingrp1+KEYBD_CLK_PIN_BYTE])
    if (cfg_boot_mode & BOOT_MODE_STDIN_MASK) == BOOT_MODE_STDIN_KEYBD
      bc.setConsole(bc#BIOS_STDIN,  bc#BIOS_KEYBOARD)
{
#endif
}

PRI boot | sbuf[100]

' if (cfg_boot_mode & BOOT_MODE_DHCP)
' if (cfg_boot_mode & BOOT_MODE_NETBOOT)
  ' try to netboot
  
  if (cfg_boot_mode & BOOT_MODE_DEBUG)
    ' load sdcard boot program   

{
#ifdef USE_SDCARD
}
  if byte[@cfg_pingrp1+SDCARD_BASEPIN_BYTE] <> "n"
    if (cfg_boot_mode & BOOT_MODE_DEBUG)
      print(string(CR,CR,"Mounting SD Card ... ")) 
    sdcard.mount(byte[@cfg_pingrp1+SDCARD_BASEPIN_BYTE])
    if (cfg_boot_mode & BOOT_MODE_DEBUG)
      puts(string("SD Card Mounted.",CR,CR,"Volume Directory:")) 
      sdcard.opendir
      repeat while 0 == sdcard.nextfile(@sbuf)
        printf(string(" %s\n"),parm(@sbuf))
        
    if strsize(@cfg_sdboot_fname) == 0
      puts(string(CR,"No SD boot filename. Boot Aborted."))

    else

      if (cfg_boot_mode & BOOT_MODE_BOOTC) ' boot a c program

      else                                 ' boot a spin program

{
#endif
}
  

PRI biosConfig(force) | len, input, addr, data, str, adj, sbuf[100]


  input := 0
  if force <> true
    repeat 3
      waitcnt(clkfreq+cnt)
      input := fds.rxcheck
      if input == 3 or input == $1B
        quit

  getEEconfig(@cfg_size, cfg_size+4)

  ' if config is blank, load defaults
  if cfg_crc16 == 0
    newEEconfig
    
  dumpEEconfigInfo(0)

  if force <> true
    if input <= 0
      return
        
  adj := bc#BIOS_DBASE_ADDR - @cfg_size
  addr := @cfg_string_data
  data := 0
  str  := 0

  repeat
    print(string(CR,"Config> "))
    input := getStdin
    if input <> CR
      puts(cstr(input))
      
    case input
      "a":
        addr := getAddr(@cbuf)-adj 
      "d": str  := 0
        data := getData(@cbuf) 
      "s":
        str  := getString(@sbuf,100,CR)
      "r":
        showField(addr)
      "w":
        setField(addr,str,data)
      "B":
        quit 
      "C":
        dumpEEconfigInfo(true) ' use guided config
      "D":
        dumpEEconfigBytes(@cfg_size, @cfg_crc16-@cfg_size+4)
        dumpEEconfigInfo(false) ' just dump
      "E":
        len := (@cfg_crc16-@cfg_size+4)
        printf(string("Erasing Config Space %d bytes\n"),parm(len))
        bytefill(@cfg_size, 0, len)
        dumpEEconfigBytes(@cfg_size, len)
        saveEEconfig(@cfg_size, @cfg_crc16-@cfg_size+4)
      "L":
        getEEconfig(@cfg_size, @cfg_crc16-@cfg_size+4)
      "N":
        newEEconfig
      "S":
        cfg_crc16 := crc.get(@cfg_size, @cfg_crc16-@cfg_size, crc#CRC_INIT)
        printf(string("\nSaving config with CRC $%04x"),parm(cfg_crc16))
        'printStdout(sconv.itoab(cfg_crc16,@cbuf,16))
        saveEEconfig(@cfg_size, @cfg_crc16-@cfg_size+4)
      "q":
        print(string(CR,"Save [y/n] ? "))
        if fds.rx == "y"
          saveEEconfig(@cfg_size, @cfg_crc16-@cfg_size+4)
        reboot  
        quit
      "?":
{
#ifdef USE_CONSOLE_CONFIG
}      
        print(string(CR,"a: set address for data read/write,"))
        print(string(CR,"s: set string data, n: set number"))
        print(string(CR,"r: read one RAM address,"))
        print(string(CR,"w: write one RAM address, "))
        print(string(CR,"C: start config utility, B: to boot,"))
        print(string(CR,"D: dumps RAM data and config fields,"))
        print(string(CR,"E: erases EEPROM space, L: loads EEPROM,"))
        print(string(CR,"N: make new config in RAM,"))
        print(string(CR,"S: save RAM config to EEPROM,"))
        print(string(CR,"?: show help and addr/data settings,"))
        print(string(CR,"q: will reboot after saving settings. "))
{
#endif
}
        puts(string(CR,CR,"Current r/w command settings:"))
        printf(string("  addr $%04x, "),parm(addr+adj))
        if strsize(str)
          printf(string("str  '%s'\n"),parm(str))
        else
          printf(string("data $%08x\n"),parm(data))
          
VAR byte cstrbuf[2]
PRI cstr(ch)
  byte[@cstrbuf+0] := ch
  byte[@cstrbuf+1] := 0
  return @cstrbuf
          
PRI getAddr(buf) | ch, ii
  ii := 0
  repeat 8
    ch := fds.rx
    if ch == CR
      quit
    byte[buf+ii] := ch
    ii++
    byte[buf+ii] := 0
    putchar(ch)
  return sconv.atoib(buf,16)
  
PRI getData(buf) | ch, ii
  ii := 0
  repeat 8
    ch := fds.rx
    if ch == CR
      quit
    byte[buf+ii] := ch
    ii++
    byte[buf+ii] := 0
    putchar(ch)
  return sconv.atoib(buf,16)

PRI getString(buf,lim,trm) | ch, ii
  ii := 0
  byte[buf] := 0
  repeat
    ch := fds.rx
    if ch == trm
      quit
    byte[buf+ii] := ch
    ii++
    byte[buf+ii] := 0
    putchar(ch)
    if ii => lim
      quit
  return buf

PRI showField(addr)
  if addr
    if strsize(addr)
      print(addr)
    else
      print(sconv.itoabx(long[addr],@cbuf,16,8))
   
PRI setField(addr,str,data) | ii 
  if addr
    if str
      ii := 0
      repeat
        byte[addr+ii] := byte[str+ii]
        ii++
      while(byte[str+ii])
    else
      long[addr] := data
   
PRI getEEconfig(dptr,length) | mycrc, ii, rc
' read database

  printf(string(CR,"Loading EEPROM config ... %d bytes"),parm(length))

  eep.Initialize
  repeat ii from 0 to length-1 step eep#PageSize
    rc := eep.ReadPage(eep#EEPROM, bc#BIOS_DBASE_ADDR+ii, dptr+ii, eep#PageSize)
    if rc
      printf(string(CR,"ReadPage Error $d. Addr $%x\n"),parm(rc)+parm(bc#BIOS_DBASE_ADDR+ii))
      return rc
    
  ' start with -1 to detect blank array
  mycrc := crc.get(@cfg_size, @cfg_crc16-@cfg_size, crc#CRC_INIT)

  if(long[dptr+length-1] <> mycrc)
    printf(string(CR,"Bad CRC: Got $%04x. Expected $%04x\n"),parm(long[dptr+length-1])+parm(mycrc))

  'dumpEEconfigBytes(dptr, length)
  
  if(long[dptr+length-1] <> mycrc)
    return -1
  return 0    

PRI saveEEconfig(dptr,length) | rc, ii
' write database

  printf(string(CR,"Saving EEPROM config ... %d bytes"),parm(length))

  eep.Initialize
  repeat ii from 0 to length-1 step eep#PageSize
    rc := eep.WritePage(eep#EEPROM, bc#BIOS_DBASE_ADDR+ii, dptr+ii, eep#PageSize)
    repeat while eep.WriteWait(eep#EEPROM, bc#BIOS_DBASE_ADDR+ii)
    if rc
      printf(string(CR,"WritePage Error $d. Addr $%x\n"),parm(rc)+parm(bc#BIOS_DBASE_ADDR+ii))
      return rc
    
  'getEEconfig(dptr,length)
  
  return 0    

PRI dumpEEconfigBytes(dptr, length) | ii
  'tv.out(CR)
  repeat ii from 0 to length-1 step 4
    ifnot ii // 16
      'printStdout(string(CR,"EE@ "))
      printf(string("\n%04x "), parm(bc#BIOS_DBASE_ADDR+ii))
    print(sconv.itoabx(long[dptr+ii],@cbuf,16,8))
    ifnot ii // 16 == 15
      putchar(" ")
    'if (ii // 192) == 188
    '  repeat
  putchar($d)

{
#ifdef USE_CONSOLE_CONFIG
}
PRI configStringField(set, field, len, pnull) | ii, s, sbuf[100]
  repeat ii from 0 to len-1
    if sconv.isalnum(byte[field+ii])
      printStdout(cstr(byte[field+ii]))
    if byte[field+ii] == 0 and pnull == 0
      quit
  if(set)
    putchar(">")
    s := getString(@sbuf,len,CR)
    if strsize(s)
      setField(field,s,0)

PRI configNumberField(set, field) | s
  printStdout(sconv.itoab(long[field], @cbuf, 16))
  if(set)
    putchar(">")
    s := getData(@cbuf)
    if strsize(@cbuf)
      setField(field,0,s)

PRI configPinField(set, cname, field, pin, name) | ii, s, sbuf[10], adj
  adj := bc#BIOS_DBASE_ADDR - @cfg_size

  'printStdout(string(CR," "))
  'printStdout(name)
  'printStdout(cstr(" "))

  printf(string(CR,"%s:1:%x:%s "),parm(cname)+parm(field+pin+adj)+parm(name))
{  
  printStdout(cstr(CR))
  printStdout(cname)
  printStdout(string(":1:"))
  printStdout(sconv.itoabx(field+pin+adj,@cbuf,16,4))
  printStdout(string(": "))
  printStdout(name)
  printStdout(string(" "))
}    
  if byte[field+pin] == "n"
    putchar("n")
  elseif byte[field+pin] == "Y"
    putchar("Y")
  else
    print(sconv.itoa(byte[field+pin], @cbuf))

  if(set)
    printStdout(cstr(">"))
    s := getString(@sbuf,4,CR)
    printStdout(s)
    if strsize(@sbuf)
      case byte[s]
        "n","N":
          'printf(string("Saving n "))
          'print(sconv.itoab(field+adj+pin, @cbuf,16))
          byte[field+pin] := "n"        
          'putchar(byte[field+pin])
        "y","Y":
          'print(string("Saving Y "))
          byte[field+pin] := "Y"        
          'putchar(byte[field+pin])
        other:
          if sconv.atoi(s) => 0
            byte[field+pin] := sconv.atoi(s) 
            'print(sconv.itoab(byte[field+pin], @cbuf,16))

PRI configField(set, name, field, len, pnull) | adj
  adj := bc#BIOS_DBASE_ADDR - @cfg_size
  printf(string(CR,"%s:%d:%04x: "),parm(name)+parm(len)+parm(field+adj))
{
  printStdout(cstr(CR))
  printStdout(name)
  printStdout(string(":"))
  printStdout(sconv.itoa(len,@cbuf))
  printStdout(string(":"))
  printStdout(sconv.itoab(field+adj, @cbuf, 16))
  printStdout(string(": "))
}
  if len == 4
    configNumberField(set, field)
  else
    configStringField(set, field, len, pnull)
           
PRI configFieldPins(set, name, field, len, pnull) | adj, ii
  adj := bc#BIOS_DBASE_ADDR - @cfg_size
  printf(string(CR,"%s:%d:%04x: "),parm(name)+parm(len)+parm(field+adj))
{
  printStdout(cstr(CR))
  printStdout(name)
  printStdout(string(":"))
  printStdout(sconv.itoa(len,@cbuf))
  printStdout(string(":"))
  printStdout(sconv.itoab(field+adj, @cbuf, 16))
  printStdout(string(": "))
}
  if field == @cfg_pingrp1
    configPinField(set, name, field, TV_BASEPIN_BYTE,         @TV_BASEPIN_CFG_STR)
    configPinField(set, name, field, VGA_BASEPIN_BYTE,        @VGA_BASEPIN_CFG_STR)
    configPinField(set, name, field, NTSCPAL_BYTE,            @NTSCPAL_CFG_STR)
    configPinField(set, name, field, SDCARD_BASEPIN_BYTE,     @SDCARD_BASEPIN_CFG_STR)
    configPinField(set, name, field, MOUSE_DAT_PIN_BYTE,      @MOUSE_DAT_PIN_CFG_STR)
    configPinField(set, name, field, MOUSE_CLK_PIN_BYTE,      @MOUSE_CLK_PIN_CFG_STR)
    configPinField(set, name, field, KEYBD_DAT_PIN_BYTE,      @KEYBD_DAT_PIN_CFG_STR)
    configPinField(set, name, field, KEYBD_CLK_PIN_BYTE,      @KEYBD_CLK_PIN_CFG_STR)
  if field == @cfg_pingrp2                              
{
#ifdef USE_ENC28J60
}
    configPinField(set, name, field, ENET_CS_PIN_BYTE,        @ENET_CS_PIN_CFG_STR)
    configPinField(set, name, field, ENET_SCK_PIN_BYTE,       @ENET_SCK_PIN_CFG_STR)
    configPinField(set, name, field, ENET_SI_PIN_BYTE,        @ENET_SI_PIN_CFG_STR)
    configPinField(set, name, field, ENET_SO_PIN_BYTE,        @ENET_SO_PIN_CFG_STR)
    configPinField(set, name, field, ENET_INT_PIN_BYTE,       @ENET_INT_PIN_CFG_STR)
{
#endif
}

PRI firstIPv4byte(c) | s, fbuf[10], sbuf[10]
  putchar(c)
  fbuf[0] := c
  s := getString(@sbuf,4,".")
  bytemove(@fbuf+1,@sbuf,strsize(@sbuf))
  s := @fbuf
  putchar(".")
  return s

PRI configFieldIPv4(set, name, field, len, pnull) | adj, ii, s, c, fbuf[10], sbuf[10]
  adj := bc#BIOS_DBASE_ADDR - @cfg_size
  printf(string(CR,"%s:%d:%04x: "),parm(name)+parm(len)+parm(field+adj))
{
  printStdout(cstr(CR))
  printStdout(name)
  printStdout(cstr(":"))
  printStdout(sconv.itoa(len,@cbuf))
  printStdout(cstr(":"))
  printStdout(sconv.itoab(field+adj, @cbuf, 16))
  printStdout(string(": "))
}
  repeat ii from 0 to 4-1     ' IPv4 addrs are always 4 bytes
    printStdout(sconv.itoab(byte[field+ii],@cbuf,10))
    if ii < 3
      putchar(".")
  if set
    putchar(">")
    c := fds.rx
    if c <> LF and c <> CR
      s := firstIPv4byte(c)
      if strsize(s)
        byte[field+0] := sconv.atoi(s)
      repeat ii from 1 to 4-1     ' IPv4 addrs are always 4 bytes
        if ii < 3
          s := getString(@sbuf,4,".")
          putchar(".")
        else
          s := getString(@sbuf,4,CR)
        if strsize(s)
          byte[field+ii] := sconv.atoi(s)
   
PRI configFieldMACADDR(set, name, field, len, pnull) | adj, ii, s, c, fbuf[10], sbuf[10]
  adj := bc#BIOS_DBASE_ADDR - @cfg_size
  printf(string(CR,"%s:%d:%04x: "),parm(name)+parm(len)+parm(field+adj))
{
  printStdout(cstr(CR))
  printStdout(name)
  printStdout(cstr(":"))
  printStdout(sconv.itoa(len,@cbuf))
  printStdout(cstr(":"))
  printStdout(sconv.itoab(field+adj, @cbuf, 16))
  printStdout(string(": "))
}
  repeat ii from 0 to 6-1     ' mac addrs are always 6 bytes
    printStdout(sconv.itoabx(byte[field+ii],@cbuf,16,2))
    if ii < 5
      putchar(".")
  if set
    putchar(">")
     c := fds.rx
    if c <> LF and c <> CR
      s := firstIPv4byte(c)
      if strsize(s)
        byte[field+0] := sconv.atoi(s)
      repeat ii from 1 to 6-1     ' mac addrs are always 6 bytes
        if ii < 5
          s := getString(@sbuf,4,".")
          putchar(".")
        else
          s := getString(@sbuf,4,CR)
        if strsize(s)
          byte[field+ii] := sconv.atoib(s,16)
    
CON CFG_RDONLY = 0

PRI dumpEEconfigInfo(set)  '| ii, adj, s, sbuf[100]
  'adj := bc#BIOS_DBASE_ADDR - @cfg_size
  configField(CFG_RDONLY,string("cfg_size"),@cfg_size,@cfg_version-@cfg_size,0)
  configField(CFG_RDONLY,string("cfg_version"),@cfg_version,@cfg_chipid-@cfg_version,0)
  configField(CFG_RDONLY,string("cfg_chipid"),@cfg_chipid,@cfg_boot_mode-@cfg_chipid,0)
  configField(set,string("cfg_boot_mode"),@cfg_boot_mode, @cfg_pingrp1-@cfg_boot_mode, 0)
  configFieldPins(set,string("cfg_pingrp1"),@cfg_pingrp1, @cfg_pingrp2-@cfg_pingrp1, 0)
  configFieldPins(set,string("cfg_pingrp2"),@cfg_pingrp2, @cfg_pingrpRSV-@cfg_pingrp2, 0)
  configField(set,string("cfg_bootOk_cnt"),@cfg_bootOk_cnt, @cfg_bootFail_cnt-@cfg_bootOk_cnt, 0)
  configField(set,string("cfg_bootFail_cnt"),@cfg_bootFail_cnt, @cfg_bootCfgErr_cnt-@cfg_bootFail_cnt, 0)
  configField(set,string("cfg_bootCfgErr_cnt"),@cfg_bootCfgErr_cnt, @cfg_bootMisc_cnt-@cfg_bootCfgErr_cnt, 0)
  configField(set,string("cfg_bootMisc_cnt"),@cfg_bootMisc_cnt, @cfg_prodname-@cfg_bootMisc_cnt, 0)
  configField(set,string("cfg_prodname"),@cfg_prodname, @cfg_userid-@cfg_prodname, 0)
  configField(set,string("cfg_userid"),@cfg_userid, @cfg_passwd-@cfg_userid, 0)
  configField(set,string("cfg_passwd"),@cfg_passwd, @cfg_hostid-@cfg_passwd, 0)
  configField(set,string("cfg_hostid"),@cfg_hostid, @cfg_ipv4_address-@cfg_hostid, 0)
  configFieldIPv4(set,string("cfg_ipv4_address"),@cfg_ipv4_address, @cfg_ipv4_netmask-@cfg_ipv4_address, 0)
  configFieldIPv4(set,string("cfg_ipv4_netmask"),@cfg_ipv4_netmask, @cfg_ipv4_gateway-@cfg_ipv4_netmask,0)
  configFieldIPv4(set,string("cfg_ipv4_gateway"),@cfg_ipv4_gateway, @cfg_ipv4_dns-@cfg_ipv4_gateway,0)
  configFieldIPv4(set,string("cfg_ipv4_dns"),@cfg_ipv4_dns, @cfg_macaddr-@cfg_ipv4_dns,0)
  configFieldMACADDR(set,string("cfg_macaddr"),@cfg_macaddr, @cfg_ipv4_vlan-@cfg_macaddr,0)
  configField(set,string("cfg_ipv4_vlan"),@cfg_ipv4_vlan, @cfg_bootp_fname-@cfg_ipv4_vlan,0)
  configField(set,string("cfg_bootp_fname"),@cfg_bootp_fname, @cfg_sdboot_fname-@cfg_bootp_fname,0)
  configField(set,string("cfg_sdboot_fname"),@cfg_sdboot_fname, @cfg_string_data-@cfg_sdboot_fname,0)
  configField(set,string("cfg_string_data"),@cfg_string_data, @cfg_crc16-@cfg_string_data, true)
  cfg_crc16 := crc.get(@cfg_size, @cfg_crc16-@cfg_size, crc#CRC_INIT)
  configField(CFG_RDONLY,string("cfg_crc16"),@cfg_crc16, 4, 0)
{
#endif
}

{
#endif
}


PRI SetClk
{{
SetClk would attempt to use PinDefs.spin clock info,
but it is not what is expected in run-time.
}}
{
  result := byte[4] | $40
  byte[4] := result
  waitcnt(clkfreq/10+cnt)
  mode := pdef#_clkmode    ' this gives a 2 byte word. not at all what we want
  freq := pdef#_xinfreq
  clkset(clkmode, freq)
  clkset(mode, freq)
}

PRI putchar(c)
  print(cstr(c))

PRI puts(str)
  printStdout(str)
  printStdout(string(CR))

PRI print(str)
  printStdout(str)
  
PRI printf(fmt, arg) | pb[256]
  sf.snprintf(@pb,255,fmt,arg)
  printStdout(@pb)

PRI parm(v)
  return sf.p(v)
  
PRI printStdout(arg1)
{
#ifdef USE_TV
}            
  if bc.getConsole(bc#BIOS_STDOUT) == bc#BIOS_TV
    result := tv.str(arg1)
{
#endif
}
{
#ifdef USE_VGA
}            
  if bc.getConsole(bc#BIOS_STDOUT) == bc#BIOS_VGA
    result := vga.str(arg1)
{
#endif
}
{
#ifdef USE_FDSERIAL
}            
  if bc.getConsole(bc#BIOS_STDOUT) == bc#BIOS_FDSERIAL
    'tv.str(arg1)
    result := fds.str(arg1)
{
#endif
}

PRI printStderr(arg1)
{
#ifdef USE_TV
}            
  if bc.getConsole(bc#BIOS_STDERR) == bc#BIOS_TV
    result := tv.str(arg1)
{
#endif
}
{
#ifdef USE_VGA
}            
  if bc.getConsole(bc#BIOS_STDERR) == bc#BIOS_VGA
    result := vga.str(arg1)
{
#endif
}
{
#ifdef USE_FDSERIAL
}            
  if bc.getConsole(bc#BIOS_STDERR) == bc#BIOS_FDSERIAL
    result := fds.str(arg1)
{
#endif
}

PRI getStdin
{
#ifdef USE_KEYBOARD
}            
  if bc.getConsole(bc#BIOS_STDIN) == bc#BIOS_KEYBOARD
    result := keybd.getkey
{
#endif
}
{
#ifdef USE_FDSERIAL
}            
  if bc.getConsole(bc#BIOS_STDIN) == bc#BIOS_FDSERIAL
    result := fds.rx
{
#endif
}

PRI setConsoleColors(arg1)
{
#ifdef USE_TV
}            
  if bc.getConsole(bc#BIOS_STDOUT) == bc#BIOS_TV
    result := tv.setcolors(arg1)
  if bc.getConsole(bc#BIOS_STDERR) == bc#BIOS_TV
    result := tv.setcolors(arg1)
{
#endif
}
{
#ifdef USE_VGA
}            
  if bc.getConsole(bc#BIOS_STDOUT) == bc#BIOS_VGA
    result := vga.setcolors(arg1)
  if bc.getConsole(bc#BIOS_STDERR) == bc#BIOS_VGA
    result := vga.setcolors(arg1)
{
#endif
}

PRI consoleOut(arg1)
{
#ifdef USE_TV
}            
  if bc.getConsole(bc#BIOS_STDOUT) == bc#BIOS_TV
    result := tv.out(arg1)
  if bc.getConsole(bc#BIOS_STDERR) == bc#BIOS_TV
    result := tv.out(arg1)
{
#endif
}
{
#ifdef USE_VGA
}            
  if bc.getConsole(bc#BIOS_STDOUT) == bc#BIOS_VGA
    result := vga.out(arg1)
  if bc.getConsole(bc#BIOS_STDERR) == bc#BIOS_VGA
    result := vga.out(arg1)
{
#endif
}
            

PRI spin | type, cmd, arg1, arg2, arg3, arg4, arg5, arg6, arg7
  arg1 := long[bc#BIOS_CMD_BASE+bc#ARG1]
  arg2 := long[bc#BIOS_CMD_BASE+bc#ARG2]
{{
  The command interpreter processes commands
}}
  repeat while long[bc#BIOS_CMD_BASE] == bc#BIOS_CONFIG_INIT

  repeat ' main loop ... single threaded BIOS calls

    if long[bc#BIOS_CMD_BASE] <> 0                         'check for command

      cmd  := long[bc#BIOS_CMD_BASE]
      type := long[bc#BIOS_CMD_BASE] & bc#BIOS_CMD_MASK
      
      arg1 := long[bc#BIOS_CMD_BASE+bc#ARG1]
      arg2 := long[bc#BIOS_CMD_BASE+bc#ARG2]
      arg3 := long[bc#BIOS_CMD_BASE+bc#ARG3]
      arg4 := long[bc#BIOS_CMD_BASE+bc#ARG4]
{
      tv.out(CR)
      tv.hex(cmd, 4)
      tv.out(" ")
      tv.hex(type,4)
      tv.out(" ")
      tv.hex(arg1,8)
      tv.out(" ")
      tv.hex(arg2,8)
      tv.out(CR)
}        
      case type '   big case statement ... using if, if, if would make slightly smaller code

        bc#BIOS_RSVD:
        
          case cmd
            bc#BIOS_CONSOLE_PRINT:
              printStdout(arg1)
            bc#BIOS_CONSOLE_ERROR:
              printStderr(arg1)
            bc#BIOS_CONSOLE_INPUT:
              getStdin
            bc#BIOS_CONSOLE_SETCOLORS:
              setConsoleColors(arg1)
            bc#BIOS_CONSOLE_OUT:
              consoleOut(arg1)
{              
{
#ifdef USE_TV
}
        bc#BIOS_TV:
          case cmd        
            bc#BIOS_TV_SETCOLORS  : tv.setcolors(arg1)
            bc#BIOS_TV_OUT        : tv.out(arg1)
            bc#BIOS_TV_STR        : tv.str(arg1)
{
            bc#BIOS_TV_BIN        : tv.bin(arg1, arg2)
            bc#BIOS_TV_DEC        : tv.dec(arg1)
            bc#BIOS_TV_HEX        : tv.hex(arg1, arg2)
            bc#BIOS_TV_PRINT      : tv.str(arg1)
            bc#BIOS_TV_PUTS       : tv.str(arg1)
                                    tv.out(CR)
}
{
#endif
}
{
#ifdef USE_VGA
}
        bc#BIOS_VGA:
          case cmd
            bc#BIOS_VGA_SETCOLORS : vga.setcolors(arg1)
            bc#BIOS_VGA_OUT       : vga.out(arg1)
            bc#BIOS_VGA_STR       : vga.str(arg1)
{
            bc#BIOS_VGA_BIN       : vga.bin(arg1, arg2)
            bc#BIOS_VGA_DEC       : vga.dec(arg1)
            bc#BIOS_VGA_HEX       : vga.hex(arg1, arg2)
            bc#BIOS_VGA_PRINT     : vga.str(arg1)
            bc#BIOS_VGA_STR       : vga.str(arg1)
            bc#BIOS_VGA_PUTS      : vga.str(arg1)
                                    vga.out(CR)
}
{
#endif
}
}

{
#ifdef USE_SERIAL
}
        bc#BIOS_FDSERIAL:
          case cmd
            bc#BIOS_FDSERIAL_MODEBAUD:           ' start(rxpin, txpin, mode, baudrate) : cog
              result := fds.start(pdef#SERIAL1_RX, pdef#SERIAL1_TX, arg1, arg2)
            bc#BIOS_FDSERIAL_RXFLUSH:         ' rxflush command
              result := fds.rxflush
            bc#BIOS_FDSERIAL_RXCHECK:         ' rxcheck : rxbyte with return code in ARG1
              result := fds.rxcheck
            bc#BIOS_FDSERIAL_RXTIME:          ' rxtime(ms) : rxbyte | t
              result := fds.rxtime(arg1)
            bc#BIOS_FDSERIAL_RX:              ' rx : rxbyte
              result := fds.rx
            bc#BIOS_FDSERIAL_TX:              ' tx(txbyte)
              result := fds.tx(arg1)
            bc#BIOS_FDSERIAL_STR:             ' str(stringptr)
              result := fds.str(arg1)
{
            bc#BIOS_FDSERIAL_DEC:             ' dec(value) | i
              result := fds.dec(arg1)
            bc#BIOS_FDSERIAL_HEX:             ' hex(value, digits)
              fds.hex(arg1, arg2)
            bc#BIOS_FDSERIAL_BIN:             ' bin(value, digits)
              fds.bin(arg1, arg2)
}              
{
#endif
}
{
#ifdef USE_MOUSE
}
        bc#BIOS_MOUSE:
          case cmd
            bc#BIOS_MOUSE_PRESENT:            ' present : type
              result := mouse.present
            bc#BIOS_MOUSE_BUTTON:             ' button(b) : state
              result := mouse.button(arg1)
            bc#BIOS_MOUSE_BUTTONS:            ' buttons : states
              result := mouse.buttons
            bc#BIOS_MOUSE_ABS_X:              ' abs_x : x
              result := mouse.abs_x
            bc#BIOS_MOUSE_ABS_Y:              ' abs_y : y
              result := mouse.abs_y
            bc#BIOS_MOUSE_ABS_Z:              ' abs_z : z
              result := mouse.abs_z
            bc#BIOS_MOUSE_DELTA_RESET:        ' delta_reset
              result := mouse.delta_reset
            bc#BIOS_MOUSE_DELTA_X:            ' delta_x : x | newx
              result := mouse.delta_x
            bc#BIOS_MOUSE_DELTA_Y:            ' delta_y : y | newy
              result := mouse.delta_y
            bc#BIOS_MOUSE_DELTA_Z:            ' delta_z : z | newz
              result := mouse.delta_z
            bc#BIOS_MOUSE_BOUND_LIMITS:       ' bound_limits(xmin, ymin, zmin, xmax, ymax, zmax) | i
              result := mouse.bound_limits(arg1, arg2, arg3, arg4, long[bc#BIOS_CMD_BASE+bc#ARG5], long[bc#BIOS_CMD_BASE+bc#ARG6])
            bc#BIOS_MOUSE_BOUND_SCALES:       ' bound_scales(x_scale, y_scale, z_scale)
              result := mouse.bound_scales(arg1, arg2, arg3)
            bc#BIOS_MOUSE_BOUND_PRESET:       ' bound_preset(x, y, z) | i, d
              result := mouse.bound_preset(arg1, arg2, arg3)
            bc#BIOS_MOUSE_BOUND_X:            ' bound_x : x
              result := mouse.bound_x
            bc#BIOS_MOUSE_BOUND_Y:            ' bound_y : y
              result := mouse.bound_y
            bc#BIOS_MOUSE_BOUND_Z:            ' bound_z : z
              result := mouse.bound_z
{
#endif
}

{
#ifdef USE_KEYBOARD
}
        bc#BIOS_KEYBOARD:
          case cmd
            bc#BIOS_KEYBOARD_STARTX:          ' startx(dpin, cpin, locks, auto) : okay
              result := keybd.startx(arg1, arg2, arg3, arg4)
            bc#BIOS_KEYBOARD_PRESENT:         ' present : truefalse
              result := keybd.present
            bc#BIOS_KEYBOARD_KEY:             ' key : keycode
              result := keybd.key
            bc#BIOS_KEYBOARD_GETKEY:          ' getkey : keycode
              result := keybd.getkey
            bc#BIOS_KEYBOARD_NEWKEY:          ' newkey : keycode
              result := keybd.newkey
            bc#BIOS_KEYBOARD_GOTKEY:          ' gotkey : truefalse
              result := keybd.gotkey
            bc#BIOS_KEYBOARD_CLEARKEYS:       ' clearkeys
              keybd.clearkeys
            bc#BIOS_KEYBOARD_KEYSTATE:        ' keystate(k) : state  
              result := keybd.keystate(arg1)
{
#endif
}

{
#ifdef USE_EEPROM
}
        bc#BIOS_EEPROM:
          case cmd
            bc#BIOS_EEPROM_INIT:              ' Initialize()
              result := eep.Initialize
            bc#BIOS_EEPROM_WRITE:             ' Write(data) : ackbit
              result := eep.Write(arg1)
            bc#BIOS_EEPROM_READ:              ' Read(ackbit): data
              result := eep.Read(arg1)
            bc#BIOS_EEPROM_READPAGE:          ' ReadPage(devSel, addrReg, dataPtr, count) : ackbit
              result := eep.ReadPage(arg1, arg2, arg3, arg4)
            bc#BIOS_EEPROM_READBYTE:          ' ReadByte(devSel, addrReg) : data
              result := eep.ReadByte(arg1, arg2)
            bc#BIOS_EEPROM_READWORD:          ' ReadWord(devSel, addrReg) : data
              result := eep.ReadWord(arg1, arg2)
            bc#BIOS_EEPROM_READLONG:          ' ReadLong(devSel, addrReg) : data
              result := eep.ReadLong(arg1, arg2)
            bc#BIOS_EEPROM_WRITEPAGE:         ' WritePage(devSel, addrReg, dataPtr, count) : ackbit
              result := eep.WritePage(arg1, arg2, arg3, arg4)
            bc#BIOS_EEPROM_WRITEBYTE:         ' WriteByte(devSel, addrReg, data)
              result := eep.WriteByte(arg1, arg2, arg3)
            bc#BIOS_EEPROM_WRITEWORD:         ' WriteWord(devSel, addrReg, data)
              result := eep.WriteWord(arg1, arg2, arg3)
            bc#BIOS_EEPROM_WRITELONG:         ' WriteLong(devSel, addrReg, data)
              result := eep.WriteLong(arg1, arg2, arg3)
            bc#BIOS_EEPROM_WRITEWAIT:         ' WriteWait(devSel, addrReg) : ackbit  
              result := eep.WriteWait(arg1, arg2)
{
#endif
}

{
#ifdef USE_SDCARD
}
        bc#BIOS_SDCARD:
          if byte[@cfg_pingrp1+SDCARD_BASEPIN_BYTE] <> "n"
            case cmd
              bc#BIOS_SDCARD_MOUNT:             ' mount(basepin)
                result := sdcard.mount(byte[@cfg_pingrp1+SDCARD_BASEPIN_BYTE])
                'result := sdcard.mount(pdef#FSRW1_DO) ' mount can throw an exception, so best to call with \abort syntax
              bc#BIOS_SDCARD_PFLUSH:            ' pflush : rc
                result := sdcard.pflush
              bc#BIOS_SDCARD_PCLOSE:            ' pclose : rc
                result := sdcard.pclose
              bc#BIOS_SDCARD_POPEN:             ' popen(s, mode) : rc
                result := sdcard.popen(arg1, arg2)
              bc#BIOS_SDCARD_PREAD:             ' pread(ubuf, count) : rc
                result := sdcard.pread(arg1, arg2)
              bc#BIOS_SDCARD_PGETC:             ' pgetc : rc
                result := sdcard.pgetc
              bc#BIOS_SDCARD_PWRITE:            ' pwrite(ubuf, count) : rc
                result := sdcard.pwrite(arg1, arg2)
              bc#BIOS_SDCARD_PPUTC:             ' pputc(c) : rc
                result := sdcard.pputc(arg1)
              bc#BIOS_SDCARD_OPENDIR:           ' opendir
                result := sdcard.opendir
              bc#BIOS_SDCARD_NEXTFILE:          ' nextfile(fbuf) : rc
                result := sdcard.nextfile(arg1)        
{
#endif
}

        other:

      long[bc#BIOS_CMD_BASE+bc#ARG1] := result          ' return code
      long[bc#BIOS_CMD_BASE] := 0 'let caller know command is done

      'repeat

  